# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

#################
# Configuration #
#################

proc default_config {} {
  # set up default config
  lappend conf Audio,Samplerate 48000
  lappend conf Audio,Input,Device {audio:-1}
  lappend conf Audio,Output,Device {audio:-1}
  lappend conf Audio,Bitdepth 16
  # expected soundcard latency in ms
  # used for computing the total record length
  lappend conf Audio,Latency 200

  # The output channel and level definition
  lappend conf Output,Channel Left
  # output RMS level
  # The output level wrt full scale is computed by:
  # lvl = sqrt(2)*Output,Level / (Output,{Global|Left|Right},Gain * Output,{Global|Left|Right},Sensitivity)
  # Actually, the program assumes a test signal with 3 dB crest factor. As this is mostly true for sweeps, 
  # pseudo-random noises have usually a crest factor of 10 to 12 dB.
  lappend conf Output,Level 2.83
  # maximum peak FS level to prevent nasty clipping
  lappend conf Output,MaxFS 0.99
  # gain of external amplifier
  lappend conf Output,Global,Gain 1.0
  # sensitivity of soundcard output in Vrms/FS
  lappend conf Output,Global,Sensitivity 1.0
  lappend conf Output,Left,Gain 1.0
  lappend conf Output,Left,Sensitivity 1.0
  lappend conf Output,Right,Gain 1.0
  lappend conf Output,Right,Sensitivity 1.0

  # DUT channel for single and dual channel measurements
  lappend conf Input,Channel Left
  # Reference channel for dual channel measurements
  lappend conf Input,Reference Right
  # Gain of external amplifier
  lappend conf Input,Global,Gain 1.0
  # Sensitivity of soundcard input in Vrms/FS
  lappend conf Input,Global,Sensitivity 1.0
  lappend conf Input,Left,Gain 1.0
  lappend conf Input,Left,Sensitivity 1.0
  lappend conf Input,Right,Gain 1.0
  lappend conf Input,Right,Sensitivity 1.0

  # list of defined microphones
  lappend conf Mic,List {1}
  lappend conf Mic,Default 1
  # gain of microphone amplifier
  lappend conf Mic,Global,Gain 1.0
  # Microphone sensitivity in mV/Pa
  lappend conf Mic,Global,Sensitivity 1.0
  # Microphone distance.
  # Used for automatic gate setting in dual channel mode
  lappend conf Mic,Global,Distance -1
  lappend conf Mic,1,Gain 1.0
  lappend conf Mic,1,Sensitivity 1.0
  lappend conf Mic,1,Distance -1
  lappend conf Mic,1,Compensation {{} nosafe}
  lappend conf Mic,1,Compensation,File {}

  lappend conf Signal,Type logsweep
  lappend conf Signal,Length 2000
  lappend conf Signal,Spectrum pink
  lappend conf Signal,Bandwidth {{20 20e3}}
  lappend conf Signal,External,Reference {./rc/reference.dat}

  lappend conf Measure,Mode {dual}
  lappend conf Measure,External,Trigger,Level -36
  lappend conf Measure,External,Trigger,Type RMS

  # default gate width
  # shifted by microphone distance
  lappend conf Processing,Gate {{0 100}}
  # Automatic or manual gate setting
  # In single channel measurements the "auto" mode will set the
  # gate start 10 ms ahead of the IR maximum. In dual channel measurements, it will auto correct
  # for the microphone distance. The length of the gate will not be affected.
  # In manual mode, the gate can only be modified via the IR window and the cursors.
  lappend conf Processing,Gate,Mode auto
  # type and width of gating window
  # these are sane defaults
  lappend conf Processing,Gate,Window,Type hann
  # gating width in %
  lappend conf Processing,Gate,Window,Width 25
  # Method to calculate the frequency response. Currently only the FFT is implemented, but
  # other methods, based on statistical system analysis (e.g. Prony's method), may be in the future.
  lappend conf Processing,FR,Method FFT
  # FFT size. Use 0 for automatic selection
  # otherwise the number of samples must be large enough so that the whole gated IR
  # fits in the FFT array
  lappend conf Processing,FFT,Size 0
  # Smoothing of the frequency response in 1/n octave
  lappend conf Processing,FFT,Smoothing 6
  # list of harmonic distortion components to show
  lappend conf Processing,HD,List {{2 3 5}}
  # Standard mode is "compressed". Then the harmonic distortion curves are frequency compressed,
  # i.e. the frequency axis for the n'th harmonic distortion is scaled by 1/n. This is the usual way
  # other programs display the distortion.
  # When mode is set to "uncompressed" the frequency axis is not scaled, thus there might be a better
  # correlation between non-linear distortion components and a possible gain by strong linear distortion
  # as often seen in single speaker drivers.
  lappend conf Processing,HD,Mode compressed
  # Weighting of distortion components
  # none: no weighting (standard)
  # linear: the n'th HD component is weighted by n/2
  # quadratic: the n'th HD component is weighted by n^2/4
  # A/B/C/D/ITU468: each HD component is weighted according to these well known weighting curves
  #
  # Recent data (e.g. Geddes) supports that harmonic distortion of low order is nearly imperceivable,
  # and the higher the order the more perceivable they become. Hence the use of no weighting for the distortions
  # does not correlate with the perception of our auditory system.
  # The different weightings try to cope with that.
  #
  # This weighting is always processed before compressing the frequency axis.
  # CAVEATS:
  # - the weighting filters are just approximations of the real weighting functions, but should work nonetheless.
  lappend conf Processing,HD,Weighting none
  # Shifting of the distortion components. Often used to have a higher vertical resolution. Positive value.
  lappend conf Processing,HD,Shifting 0
  # If >0, a copy of the frequency response is displayed shifted down this value
  # This gives a visible distortion limit.
  # A special value is "ISO532B". It will show a limit curve for each component in Processing,HD,List based on
  # the masking effects derived by Zwicker for loudness computation. This feature is currently not implemented, but certainly
  # will in the future and is surely of great use in evaluating a speakers performance.
  lappend conf Processing,HD,FRShift 0

  # If empty, the burst decay will simply normalize the time axis to periods.
  # If "peaks", then the peaks of each evaluated frequency will be aligned with the peak of the reference
  lappend conf Processing,BD,Align {peaks}

  # DISPLAY section
  # "absolute" or "relative" scaling
  # "absolute" scaling is only possible with single channel measurements
  # Note that the measured impulse response itself is always scaled to 1V. These variables
  # only influence the display, not the actual measurement. 
  # If measured single channel, "relative" assumes the reference voltage given below. 
  # Thus, the displayed curve will be completely wrong if the amplifiers gain wasn't set to 
  # the values in the "Output" section
  lappend conf Display,Scaling relative
  # Reference level
  lappend conf Display,Scaling,Reference 2.83
  # frequency response vertical display range (dB)
  lappend conf Display,FR,VRange 50
  # frequency response horizontal display range (Hz, 2 item list or "auto")
  lappend conf Display,FR,HRange {auto}

  # distortion frequency response vertical display range
  lappend conf Display,HD,VRange 80
  # distortion frequency response horizontal display range (Hz, 2 item list or "auto")
  lappend conf Display,HD,HRange {auto}
  
  # burst decay response vertical display range
  lappend conf Display,BD,VRange 50
  # burst decay response horizontal display range (number of Periods or "auto")
  lappend conf Display,BD,HRange {auto}
  # number of bins displayed
  lappend conf Display,Bins 200

  # OVERLAY handling
  # define if new measurements replace the current master curve or are appended at the end of the overlay list
  lappend conf Overlay,Add,Mode replace

  # Physical constants
  lappend conf Physics,C 343.0

  lappend conf System,LastDir {}

  # create ensemble command
  ::config::makeEnsemble ::setup
  # write defaults to configuration array
  setup aset $conf

  # read setup from file
  if {![load_config]} {
    ::config::writeSetup
  }
  #::config::enable_writeOnChange

  # install other traces
  setup ctrace add Audio,Samplerate write [list apply {{v1 v2 op} {
      upvar $v1 options
      audio configure Samplerate $options($v2)
      readMicCompensation
    }}]
  setup ctrace add Audio,Input,Device write [list apply {{v1 v2 op} {
      upvar $v1 options
      audio configure Input,Device $options($v2)
    }}]
  setup ctrace add Audio,Output,Device write [list apply {{v1 v2 op} {
      upvar $v1 options
      audio configure Output,Device $options($v2)
    }}]
  setup ctrace add Display,* write [list apply {{v1 v2 op} {
      upvar $v1 options
      if {$v2 eq {Display,Scaling} && $options($v2) eq {absolute}} {
        if {$options(Measure,Mode) eq {dual}} {
          catch {::cmdbar::writeToCmdBar {Absolute scaling not possible in dual channel model.} -bg red}
          set options(Display,Scaling) relative
        }
      }
    }}]
  setup ctrace add Display,FR,HRange write [list apply {{v1 v2 op} {
      upvar $v1 options
      if {$options($v2) eq {auto}} {
        set ::TkEsweepXYPlot::plots(FR,Config,X,Max) 20e3
        # autoscaling may not be possible yet, so catch it
        set ::TkEsweepXYPlot::plots(FR,Config,X,Autoscale) {on}
        catch {::TkEsweepXYPlot::autoscale FR X}
        set ::TkEsweepXYPlot::plots(FR,Config,X,Autoscale) {off}
        set ::TkEsweepXYPlot::plots(FR,Config,X,Min) 20
      } else {
        set ::TkEsweepXYPlot::plots(FR,Config,X,Min) [lindex $options($v2) 0]
        set ::TkEsweepXYPlot::plots(FR,Config,X,Max)  [lindex $options($v2) 1]
      }
      catch {::TkEsweepXYPlot::plot FR}
    }}]
  setup ctrace add Display,FR,VRange write [list apply {{v1 v2 op} {
      upvar $v1 options
      set ::TkEsweepXYPlot::plots(FR,Config,Y1,Range) $options($v2)
      catch {::TkEsweepXYPlot::plot FR}
    }}]
  setup ctrace add Display,HD,* write [list apply {{v1 v2 op} {
      upvar $v1 options
      if {$v2 eq {Display,Scaling} && $options($v2) eq {absolute}} {
        if {$options(Measure,Mode) eq {dual}} {
          catch {::cmdbar::writeToCmdBar {Absolute scaling not possible in dual channel model.} -bg red}
          set options(Display,Scaling) relative
        }
      }
    }}]

  setup ctrace add Display,HD,HRange write [list apply {{v1 v2 op} {
      upvar $v1 options
      if {$options($v2) eq {auto}} {
        set ::TkEsweepXYPlot::plots(HD,Config,X,Max) 20e3
        # autoscaling may not be possible yet, so catch it
        set ::TkEsweepXYPlot::plots(HD,Config,X,Autoscale) {on}
        catch {::TkEsweepXYPlot::autoscale HD X}
        set ::TkEsweepXYPlot::plots(HD,Config,X,Autoscale) {off}
        set ::TkEsweepXYPlot::plots(HD,Config,X,Min) 20
      } else {
        set ::TkEsweepXYPlot::plots(HD,Config,X,Min) [lindex $options($v2) 0]
        set ::TkEsweepXYPlot::plots(HD,Config,X,Max)  [lindex $options($v2) 1]
      }
      catch {::TkEsweepXYPlot::plot HD}
    }}]
  setup ctrace add Display,HD,VRange write [list apply {{v1 v2 op} {
      upvar $v1 options
      set ::TkEsweepXYPlot::plots(HD,Config,Y1,Range) $options($v2)
      catch {::TkEsweepXYPlot::plot HD}
    }}]

  setup ctrace add Display,BD,HRange write [list apply {{v1 v2 op} {
      upvar $v1 options
      if {$options($v2) eq {auto}} {
        set ::TkEsweepXYPlot::plots(BD,Config,X,Max) 20
        # autoscaling may not be possible yet, so catch it
        set ::TkEsweepXYPlot::plots(BD,Config,X,Autoscale) {on}
        catch {::TkEsweepXYPlot::autoscale BD X}
        set ::TkEsweepXYPlot::plots(BD,Config,X,Autoscale) {off}
        set ::TkEsweepXYPlot::plots(BD,Config,X,Min) 0
      } else {
        set ::TkEsweepXYPlot::plots(BD,Config,X,Max) $options($v2)
      }
      catch {::TkEsweepXYPlot::plot BD}
    }}]
  setup ctrace add Display,HD,VRange write [list apply {{v1 v2 op} {
      upvar $v1 options
      set ::TkEsweepXYPlot::plots(BD,Config,Y1,Range) $options($v2)
      catch {::TkEsweepXYPlot::plot BD}
    }}]


  setup ctrace add Processing,HD,* write [list apply {{v1 v2 op} {
      catch {updateAppTitle HD}
    }}]
  setup ctrace add Processing,BD,* write [list apply {{v1 v2 op} {
      #catch {updateAppTitle BD}
    }}]
  setup ctrace add Processing,Gate write [list apply {{v1 v2 op} {
      global master
      upvar $v1 options
      catch {::overlay::modify $::master [list Gate $options($v2)]}
    }}]
  setup ctrace add Processing,FFT,Smoothing write [list apply {{v1 v2 op} {
      global master
    }}]
  setup ctrace add Processing,* write [list apply {{v1 v2 op} {
      catch {updateAppTitle FR}
    }}]
  setup ctrace add Measure,Mode write [list apply {{v1 v2 op} {
      upvar $v1 options
      if {$options($v2) eq {dual} && $options(Display,Scaling) eq {absolute}} {
        catch {::cmdbar::writeToCmdBar {Absolute scaling not possible in dual channel mode, switched to relative scaling.} -bg yellow}
        set options(Display,Scaling) relative
      }
      if {$options($v2) eq {external}} {
        catch {::cmdbar::writeToCmdBar {Set Signal,Type to "external"} -bg yellow}
        set options(Signal,Type) external
      } else {
        if {$options(Signal,Type) eq {external}} {
          catch {::cmdbar::writeToCmdBar {Set Signal,Type to "logsweep"} -bg yellow}
          set options(Signal,Type) logsweep
        }
      }

      catch {updateAppTitle FR}
    }}]
  setup ctrace add Signal,Type write [list apply {{v1 v2 op} {
      upvar $v1 options
      if {$options($v2) eq {external}} {
        catch {::cmdbar::writeToCmdBar {Set Processing,Mode to "external"} -bg yellow}
        set options(Processing,Mode) external
      }
      catch {updateAppTitle FR}
    }}]
}

proc load_config {} {
  return [::config::readSetup strict]
}

proc readMicCompensation {{mic all}} {
  if {$mic eq {all}} {set micList [setup cget Mic,List]} else {set micList $mic}
  foreach mic $micList {
    if {[catch {set fp [open [setup cget Mic,$mic,Compensation,File] r]}]} {
      continue
    }
    set lines [split [read $fp] \n]
    close $fp

    foreach l $lines {
      if {[regexp {^[[:digit:]]} $l]} {lappend data $l}
    }
    foreach d $data break

    # split frequency, magnitude and phase
    set n [llength [split $d]]
    switch -- $n {
      2 {
        # only magnitude
        foreach d $data {
          lappend comp [lindex $d 0] [lindex $d 1] 0.0
        }
      }
      3 {
        # magnitude + phase
        foreach d $data {
          lappend comp [lindex $d 0] [lindex $d 1] [expr {$::PI*[lindex $d 2]/360}]
        }
      }
      default {
        # crap
        continue
      }
    }
    setup configure Mic,$mic,Compensation $comp
  }
}

proc expandMicCompensation {comp {Fs {}}} {
  if {[llength $comp] < 3} return {}
  if {$Fs eq {}} {
    set Fs [setup cget Audio,Samplerate]
  }
  # interpolate for linear frequency resolution
  set f [set df [lindex $comp 0]]
  # initial values at DC
  set F 0.0
  set Mag 0.0
  set Phase 0.0
  foreach {f0 m0 p0} $comp break
  foreach {f1 m1 p1} [lrange $comp 3 end] {
    while {$f < $f1} {
      lappend F $f
      lappend Mag [expr {($m1-$m0)/($f1-$f0)*($f-$f0)+$m0}]
      lappend Phase [expr {($p1-$p0)/($f1-$f0)*($f-$f0)+$p0}]
      set f [expr {$f+$df}]
    }
    set f0 $f1
    set m0 $m1
    set p0 $p1
  }
  # extrapolate until Fs/2
  # assume a 2nd order lowpass thus phase@Fs/2=-pi
  set f1 [expr {$Fs/2}]
  set p1 [expr {-$::PI}]
  while {$f < $f1} {
    lappend F $f
    lappend Mag [expr {(($m1-$m0)/($f1-$f0)*($f-$f0)+$m0)}]
    lappend Phase [expr {($p1-$p0)/($f1-$f0)*($f-$f0)+$p0}]
    set f [expr {$f+$df}]
  }

  # merge
  set comp [list]
  foreach f $F m $Mag p $Phase {
    lappend comp $f $m $p
  }
  return $comp
}

