# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

package provide audio 1.5

namespace eval ::audio {

  if {![namespace exists ::esweep]} {
    #if {[catch {load [file join [pwd] libesweeptcl.dll] esweep}]} 
    if {[catch {load libesweeptcl.so esweep}]} {
      puts stderr "Library load error: $::errorInfo"
      exit 2
    }
  }

  variable config
  array set config {
    Input,Device  audio:-1
    Output,Device  audio:-1
    Samplerate 48000
    Framesize 0
    Channels  2
    Channels,Map {Left Right}
  }

  # delete comments from array
	array unset config {#}

  # internal state variables
  variable state
  array set state {
    # {Active states are: 0 - stopped; 1 - running}
    Active 0
    # {Trigger states are: 0 - inactive; 1 - armed; 2 - fired}
    Trigger 0
    1,Level,RMS 0
    1,Level,Peak 0
    2,Level,RMS 0
    2,Level,Peak 0
  }
  # delete comments from array
	array unset state {#}

  ::trace add variable ::audio::config(Channels) write [list apply {{state name1 name2 op} {
     upvar $name1 var
     upvar $state s
     # add/reset state variables
     for {set i 1} {$i <= $var($name2)} {incr i} {
       set s($i,Level,RMS) 0
       set s($i,Level,Peak) 0
     }
     # delete spare state variables
     for {} {[info exists s($i,Level,RMS)]} {incr i} {
       array unset s $i,Level,*
     }
    }} ::audio::state]

	proc configure {opt val} {
		variable config
		if {[info exists config($opt)]} {
			set config($opt) $val
			return $val
		}
	}

	proc cget {opt} {
		variable config
		if {[info exists config($opt)]} {
			return $config($opt)
		} else {
			return -code error "$opt: no such option in configuration"
		}
	}

  proc append {opt val} {
		variable config
		if {[info exists config($opt)]} {
			lappend config($opt) $val
		}
	}

  proc ctrace {mod name ops cmdPrefix} {
    variable config
    if {[info exists config($name)]} {
      ::trace $mod variable ::audio::config($name) $ops $cmdPrefix
      return [list $ops $cmdPrefix]
    } else {
      return {}
    }
  }

	proc sconfigure {opt val} {
		variable state
		if {[info exists state($opt)]} {
			set state($opt) $val
			return $val
		}
	}

	proc sget {var} {
		variable state
		if {[info exists state($var)]} {
			return $state($var)
		} else {
			return -code error "$var: no such state variable"
		}
	}

  proc strace {mod name ops cmdPrefix} {
    variable state
    if {[info exists state($name)]} {
      ::trace $mod variable ::audio::state($name) $ops $cmdPrefix
      return [list $ops $cmdPrefix]
    } else {
      return {}
    }
  }

  namespace ensemble create -command ::audio -subcommands [list configure cget ctrace sconfigure sget strace append measure test cancel]
}

proc ::audio::map {{name {}}} {
  if {$name eq {}} {
    return [audio cget Channels,Map]
  } else {
    if {![regexp {^[+]?[[:digit:]]+$} $name]} {
      # is not an integer, try to find in Channels,Map
      return [expr {[lsearch [audio cget Channels,Map] $name]+1}]
    } else {
      # just return $name
      return $name
    }
  }
}

proc ::audio::ismapped {{name {}}} {
  if {![regexp {^[+]?[[:digit:]]+$} $name]} {
    # is not an integer, try to find in Channels,Map
    return [lsearch [audio cget Channels,Map] $name]
  } elseif {$name > 0 && [llength [audio cget Channels,Map]] >= $name} {
    return 1
  } else {return -1}
}

proc ::audio::cancel {} {
  audio sconfigure Active 0
}

proc ::audio::measure {mode args} {
  # set defaults
  set test 0
  set output {} 
  set flags {}

  # decode common options

  # is it test mode?
  if {[set i [lsearch $args -test]] >= 0} {
    set test 1
    set args [lreplace $args $i $i]
  } else {
    set test 0
  }

  foreach {opt val} $args {
    switch -glob -- $opt {
      -out* {
        # output channel
        # check for channel mapping
        foreach c $val {
          set o [::audio::map $c]
          if {$o < 1 || $o > [audio cget Channels]} {
            return -code error  "Output channel $c not mapped."
          }
          lappend output $o
        }
      }
      -sig* {-}
      -ref* {
        upvar 1 $val obj
      }
      -flags {
        set flags $val
      }
    }
  }
  if {[catch {esweep::size -obj $obj}]} {
    if {!($mode eq {external} && $test) || $mode eq {external} || $mode eq {single}} {
      return -code error {No reference signal supplied}
    } else {
      set reference {}
    }
  } else {
    if {[esweep::type -obj $obj] ne {wave} && [esweep::type -obj $obj] ne {complex}} {
      return -code error {Reference signal of wrong type supplied}
    }
    set reference [esweep::clone -src $obj]
  }
  set common [list $test $output $reference $flags]

  switch -- $mode {
    {dual} {-}
    {single} {
      # single channel measurement
      return [::audio::measure!internal $mode $common $args]
    }
    {external} {
      # external source measurement
      return [::audio::measure!external $common $args]
    }
    default {
      # unknown
      return -code error "Unknown measurement mode $mode."
    }
  }
}

proc ::audio::measure!internal {mode common args} {
  lassign $common test output reference flags
  set recLength -1
  set input 1
  set inref 2
  foreach {opt val} [join $args] {
    switch -glob -- $opt {
      -in* {
        # input channel(s)
        # check for channel mapping
        set input [::audio::map [lindex $val 0]]
        if {$input < 1 || $input > [audio cget Channels]} {
          return -code error  "Input channel [lindex $val 0] not mapped."
        }
        if {$mode eq {dual}} {
          if {[llength $val] != 2} {
            return -code error {Channel definition not complete for dual channel measurement}
          }
          set inref [::audio::map [lindex $val 1]]
          if {$inref < 1 || $inref > [audio cget Channels]} {
            return -code error "Reference channel [lindex $val 1] not mapped."
          }
        } else {
          set inref $input
        }
      }
      -rec* {
        # recording length in ms, we check the validity later
        set recLength $val
      }
    }
  }

  set type [esweep::type -obj $reference]
  set size [esweep::size -obj $reference]
  set Fs [esweep::samplerate -obj $reference]
  if {$Fs != [audio cget Samplerate]} {
    return -code error {Samplerate of reference not equal to samplerate of audio device}
  }

  if {[catch {lassign [::audio::open $output $input] au_in au_out} msg]} {return -code error $msg}
  if {[set framesize [esweep::audioQuery -handle $au_in -param framesize]] != [esweep::audioConfigure -handle $au_out -param framesize -value $framesize]} {
    esweep::audioClose -handle $au_in
    esweep::audioClose -handle $au_out
    return -code error {Input and output device have to have the same framesize}
  }

  if {!$test} {
    if {$recLength < 0} {
      return -code error {No record length given!}
    }
    if {[set recSize [expr {int(0.5+$Fs*$recLength/1000.0)}]] < $size} {
      return -code error {Record length < reference length!}
    }
  } else {
    set recSize $size
  }
  # make record size a multiple of the soundcard frame size
  set recSize [expr {$framesize*int(0.5+ceil(double($recSize)/$framesize))}]

  # generate input and output container
  for {set i 0} {$i < $input || $i < $inref} {incr i} {
    lappend in [esweep::create -type $type -size $recSize -samplerate $Fs]
  }
  set inobj [lindex $in [expr {$input-1}]]
  set inrefobj [lindex $in [expr {$inref-1}]]
  for {set i 0} {$i < [lindex [lsort -increasing -integer $output] end]} {incr i} {
    lappend out [esweep::create -type $type -size $recSize -samplerate $Fs]
  }
  # copy reference signal to all set output channels
  foreach c $output {
    set outobj [lindex $out [expr {$c-1}]]
    esweep::copy -src $reference -dst outobj
  }

  set delay [expr {int((150.0*$framesize)/$Fs)}]
  set waiter 0

  # set state running
  audio sconfigure Active 1
  esweep::audioSync -handle $au_in
  esweep::audioSync -handle $au_out

  set ioff [set ooff 0]
  while {$ioff < $recSize} {
    if {[audio sget Active]} {
      set oframe [esweep::audioOut -handle $au_out -signals $out -offset $ooff]
      set iframe [esweep::audioIn -handle $au_in -signals $in -offset $ioff]
      if {$iframe - $ioff > 0} {
        audio sconfigure $input,Level,Peak [esweep::max -obj $inobj -from $ioff -to [expr {$iframe-1}]]
        set rms [expr {[esweep::sqsum -obj $inobj -from $ioff -to [expr {$iframe-1}]]/($iframe-$ioff)}]
        if {$rms > 0.0} {set rms [expr {sqrt($rms)}]}
        audio sconfigure $input,Level,RMS $rms
        if {$mode eq {dual}} {
          audio sconfigure $inref,Level,Peak [esweep::max -obj $inrefobj -from $ioff -to [expr {$iframe-1}]]
          set rms [expr {[esweep::sqsum -obj $inrefobj -from $ioff -to [expr {$iframe-1}]]/($iframe-$ioff)}]
          if {$rms > 0.0} {set rms [expr {sqrt($rms)}]}
          audio sconfigure $inref,Level,RMS $rms
        }
      }
      set ioff $iframe
      set ooff $oframe
      after $delay [list incr waiter]
      vwait waiter
    } else {
      # break when cancelled
      break
    }
    # repeat while in test mode
    if {$test && [audio sget Active]} {
      set ioff [set ooff 0]
    }
  }
  esweep::audioSync -handle $au_in
  esweep::audioSync -handle $au_out
  esweep::audioClose -handle $au_in
  esweep::audioClose -handle $au_out

  audio sconfigure $input,Level,Peak 0
  audio sconfigure $input,Level,RMS 0
  if {$mode eq {dual}} {
    audio sconfigure $inref,Level,Peak 0
    audio sconfigure $inref,Level,RMS 0
  }
  update

  if {![audio sget Active]} {
    # measurement cancelled by external event
    return NULL
  }
  audio sconfigure Active 0

  if {[lsearch $flags {rec_only}] >= 0} {
    if {$mode eq {dual}} {
      return [list $inobj $inrefobj]
    } else {
      return [list $inobj]
    }
  } else {
    # make IR
    esweep::fft -inplace -obj inobj
    esweep::toPolar -obj inobj
    if {$mode eq {dual}} {
      esweep::fft -inplace -obj inrefobj
      esweep::toPolar -obj inrefobj
      esweep::expr inobj / $inrefobj
    } else {
      foreach c $output {
        set outobj [lindex $out [expr {$c-1}]]
        break
      }
      esweep::fft -inplace -obj outobj
      esweep::toPolar -obj outobj
      esweep::expr inobj / $outobj
    }
    esweep::expr inobj / [esweep::size -obj $inobj]
    esweep::ifft -inplace -obj inobj

    # Done
    return $inobj
  }
}

proc ::audio::measure!external {common args} {
  lassign $common test output reference
  # set defaults
  set trigType Peak
  set trigLevel 0.5
  set trigFrames 10
  set input 1
  set inref 2
  set mode single
  set wait 100

  foreach {opt val} [join $args] {
    switch -glob -- $opt {
      -in* {
        # input channel
        # check for channel mapping
        if {[llength $val] == 1} {
          set mode single
        } elseif {[llength $val] == 2} {
          set mode dual
        } else {
          return -code error  {Too many input channels specified}
        }
        set input [::audio::map [lindex $val 0]]
        if {$input < 1 || $input > [audio cget Channels]} {
          return -code error  "Input channel [lindex $val 0] not mapped."
        }
        if {$mode eq {dual}} {
          set inref [::audio::map [lindex $val 1]]
          if {$inref < 1 || $inref > [audio cget Channels]} {
            return -code error  "Reference channel [lindex $val 1] not mapped."
          }
          if {$inref == $input} {
            return -code error  "Input channel $input and reference channel $inref must be different."
          }
        }
      }
      -trig* {
        # trigger measurement at {RMS/Peak} level...
        if {[llength $val] == 1} {
          set trigLevel $val
        } elseif {[llength $val] > 1} {
          # first element in list gives the level type
          switch -- [lindex $val 0] {
            RMS -
            Peak {
              set trigType [lindex $val 0]
              set trigLevel [lindex $val 1]
            }
            default {
              return -code error "Unknown trigger type [lindex $val 0]"
            }
          }
          if {[llength $val] == 3} {
            set trigFrames [lindex $val 2]
          }
        }
        if {[catch {expr {int($trigLevel)}}]} {
          return -code error "Trigger level [lindex $val 1] not a number"
        }
        if {$trigLevel < 0} {
          return -code error {Trigger level must be positive}
        }
        if {!([regexp {^[+]?[[:digit:]]+$} $trigFrames] && $trigFrames > 0)} {
          # not an integer
          return -code error "Trigger frames $trigFrames not an (positive) integer"
        }
      }
      -wait -
      -del* {
        # delay after triggering
        if {!([regexp {^[+]?[[:digit:]]+$} $val] && $val > 0)} {
          # not an integer
          return -code error "Trigger delay $val not an (positive) integer"
        }
        set wait $val
      }
    }
  }
  if {[catch {set au [lindex [::audio::open 0 [expr {$inref > $input ? $inref : $input}]] 0]} msg]} {return -code error $msg}

  if {$test} {
    set type wave
    set size [esweep::audioQuery -handle $au -param framesize]
    set Fs [esweep::audioQuery -handle $au -param samplerate]
  } else {
    if {[catch {esweep::size -obj $reference}]} {
      return -code error {No reference signal supplied}
    }
    set type [esweep::type -obj $reference]
    set size [esweep::size -obj $reference]
    set Fs [esweep::samplerate -obj $reference]
    if {$Fs != [audio cget Samplerate]} {
      esweep::audioClose -handle $au
      return -code error {Samplerate of reference signal not equal to samplerate of audio device}
    }
  }
  # create input signal container
  for {set i 0} {$i < $input || $i < $inref} {incr i} {
    lappend in [esweep::create -type $type -samplerate $Fs -size $size]
  }
  set inobj [lindex $in [expr {$input-1}]]
  if {$mode eq {single}} {
    set refobj $reference
  } else {
    set refobj [lindex $in [expr {$inref-1}]]
  }

  set delay [expr {int((250.0*[esweep::audioQuery -handle $au -param framesize])/$Fs)}]
  set waiter 0

  # set state running
  audio sconfigure Active 1
  audio sconfigure Trigger 1

  esweep::audioSync -handle $au
  # wait for trigger
  set i 0
  while {[audio sget Active] && [audio sget Trigger] == 1} {
    set frame [esweep::audioIn -handle $au -signals $in]
    if {$frame > 0} {
      audio sconfigure $input,Level,Peak [esweep::max -obj $inobj -from 0 -to [expr {$frame-1}]]
      set rms [expr {[esweep::sqsum -obj $inobj -from 0 -to [expr {$frame-1}]]/$frame}]
      if {$rms > 0.0} {set rms [expr {sqrt($rms)}]}
      audio sconfigure $input,Level,RMS $rms
      if {[audio sget $input,Level,$trigType] > $trigLevel} {
        if {[incr i] > $trigFrames} {
          audio sconfigure Trigger 2
        }
      } else {
        set i 0
      }
      if {$mode eq {dual}} {
        audio sconfigure $inref,Level,Peak [esweep::max -obj $refobj -from 0 -to [expr {$frame-1}]]
        set rms [expr {[esweep::sqsum -obj $refobj -from 0 -to [expr {$frame-1}]]/$frame}]
        if {$rms > 0.0} {set rms [expr {sqrt($rms)}]}
        audio sconfigure $inref,Level,RMS $rms
      }
    }
    after $delay [list incr waiter]
		vwait waiter
  }
  if {$test} {
    return NULL
  }
  if {[audio sget Active] && [audio sget Trigger] == 2} {
    # wait one signal period to reach steady state
    after $wait [list incr waiter]
    vwait waiter
  }
  # start measurement
  set offset 0
  while {$offset < $size} {
		if {![audio sget Active]} break
		set frame [esweep::audioIn -handle $au -signals $in -offset $offset]
    if {$frame - $offset > 0} {
      audio sconfigure $input,Level,Peak [esweep::max -obj $inobj -from $offset -to [expr {$frame-1}]]
      set rms [expr {[esweep::sqsum -obj $inobj -from $offset -to [expr {$frame-1}]]/($frame-$offset)}]
      if {$rms > 0.0} {set rms [expr {sqrt($rms)}]}
      audio sconfigure $input,Level,RMS $rms
      if {$mode eq {dual}} {
        audio sconfigure $inref,Level,Peak [esweep::max -obj $refobj -from 0 -to [expr {$frame-1}]]
        set rms [expr {[esweep::sqsum -obj $refobj -from 0 -to [expr {$frame-1}]]/$frame}]
        if {$rms > 0.0} {set rms [expr {sqrt($rms)}]}
        audio sconfigure $inref,Level,RMS $rms
      }
    }
    set offset $frame
		after $delay [list incr waiter]
		vwait waiter
	}
  esweep::audioSync -handle $au
  esweep::audioClose -handle $au
  audio sconfigure $input,Level,Peak 0
  audio sconfigure $input,Level,RMS 0
  audio sconfigure $inref,Level,Peak 0
  audio sconfigure $inref,Level,RMS 0
  update

  if {![audio sget Active]} {
    # measurement cancelled by external event
    return NULL
  }
  audio sconfigure Active 0
  update

  # make IR
  esweep::fft -inplace -obj inobj
  esweep::fft -inplace -obj refobj

  esweep::toPolar -obj inobj
  esweep::toPolar -obj refobj
  esweep::expr inobj / $refobj
  esweep::expr inobj / [esweep::size -obj $inobj]
  esweep::ifft -inplace -obj inobj
  if {$mode eq {dual}} {
    esweep::fft -inplace -obj reference
    esweep::toPolar -obj reference
    esweep::expr refobj / $reference
    esweep::expr refobj / [esweep::size -obj $refobj]
    esweep::ifft -inplace -obj refobj
  }

  # Done
  # Don't forget to rotate the IR to have the start whereever you like
  return [list $inobj $refobj]
}

proc ::audio::open {outchan inchan} {
  set au_in {}
  set au_out {}
  # first try to open the audio device in full-duplex, if necessary
  # if that fails, or we need only an output/an input device, try to open for playback or recording, respectively
  if {$outchan > 0 && $inchan > 0 && ([audio cget Output,Device] eq [audio cget Input,Device])} {
    if {[catch {set au_in [set au_out [esweep::audioOpen -device [audio cget Output,Device] -mode {duplex}]]} msg]} {
      return -code error $msg
    }
  } else {
    if {$outchan > 0} {
      if {[catch {set au_out [esweep::audioOpen -device [audio cget Output,Device] -mode {output}]} msg]} {
        return -code error $msg
      }
    }
    if {$inchan > 0} {
      if {[catch {set au_in [esweep::audioOpen -device [audio cget Input,Device] -mode {input}]} msg]} {
        return -code error $msg
      }
    }
  }

  # configure the correct channel count
  if {$outchan > 0} {
    # try to set the samplerate
    if {[catch {esweep::audioConfigure -handle $au_out -param samplerate -value [audio cget Samplerate]}]} {
      return -code error {Samplerate not supported}
    }
    # try to set the framesize
    if {[audio cget Framesize] > 0} {
      if {[catch {esweep::audioConfigure -handle $au_out -param framesize -value [audio cget Framesize]}]} {
        return -code error {Framesize not supported}
      }
    }
    if {[esweep::audioConfigure -handle $au_out -param play_channels -value [audio cget Channels]] < 0} {
      # if we can't configure it, don't panic: if the device has enough channels
      # opened to include $outchan, we'll be fine
      if {[set play_channels [esweep::audioQuery -handle $au_out -param play_channels]] < $outchan} {
        # won't work
        esweep::audioClose -handle $au_out
        return -code error {Can't open enough channels for output}
      }
    }
    esweep::audioSync -handle $au_out
  }

  if {$inchan > 0} {
    # try to set the samplerate
    if {[catch {esweep::audioConfigure -handle $au_in -param samplerate -value [audio cget Samplerate]}]} {
      return -code error {Samplerate not supported}
    }
    # try to set the framesize
    if {[audio cget Framesize] > 0} {
      if {[catch {esweep::audioConfigure -handle $au_in -param framesize -value [audio cget Framesize]}]} {
        return -code error {Framesize not supported}
      }
    }
    if {[esweep::audioConfigure -handle $au_in -param record_channels -value [audio cget Channels]] < 0} {
      # if we can't configure it, don't panic: if the device has enough channels
      # opened, we'll be fine
      if {[set rec_channels [esweep::audioQuery -handle $au_in -param record_channels]] < $inchan} {
        # won't work
        esweep::audioClose -handle $au_in
        return -code error {Can't open enough channels for input}
      }
    }
    esweep::audioSync -handle $au_in
  }

  return [list $au_in $au_out]
}

proc ::audio::resample {in FS_out {cutoff_factor 0.9} {filter_length 128}}  {
  set FS_in [esweep::samplerate -obj $in]
  if {$FS_in == $FS_out} {return $in}
  set in_size [esweep::size -obj $in]
  if {$in_size > $FS_in} {
    set factor [expr {ceil(double($in_size)/$FS_in)}]
  } else {
    set factor [expr {1.0/pow(10, floor(log10(double($FS_in)/$in_size)))}]
  }
  set in_size [expr {int(0.5+$factor*$FS_in)}]
  set out_size [expr {int(0.5+$factor*$FS_out)}]

  set res_in [esweep::create -type [esweep::type -obj $in] -samplerate $FS_in -size $in_size]
  esweep::copy -dst res_in -src $in
  esweep::toWave -obj res_in
  set res_out [esweep::create -type [esweep::type -obj $res_in] -samplerate $FS_out -size $out_size]

  if {$FS_in > $FS_out} {
    set sinc [esweep::create -type wave -samplerate $FS_in -size $filter_length]
    esweep::generate sinc -obj sinc -frequency [expr {$cutoff_factor*$FS_out/2.0}]
  } else {
    set sinc [esweep::create -type wave -samplerate $FS_out -size $filter_length]
    esweep::generate sinc -obj sinc -frequency [expr {$cutoff_factor*$FS_in/2.0}]
  }
  esweep::window -obj sinc -leftwin hann -leftwidth 50 -rightwin hann -rightwidth 50

  set filter [list $sinc NULL [esweep::create -type complex -size $filter_length -samplerate [esweep::samplerate -obj $sinc]]]
  set carry [esweep::create -type wave -size $filter_length -samplerate [esweep::samplerate -obj $sinc]]
  esweep::resample -in res_in -out res_out -carry carry -filter $filter

  return $res_out
}

